/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <MiXen.h>
#include <memory.h>
#include <string.h>
#include <systm.h>

#include <xen.h>
#include <vcpu.h>

/*
 * While I couldn't find documentation on it, reading the source it seems
 * that the VCPUs are numbered 0 to n-1 with 0 being the initially running
 * VCPU.  For now, we allocate vcpus until we get some error from
 * VCPUOP_register_vcpu_info/VCPUOP_initialise.
 */
void
find_vcpus(void)
{
    int id, vicnt = 0;
    vcpu_info_t *vi, *fvi = 0;
    vcpu_register_vcpu_info_t rvi;
    vcpu_guest_context_t gc;

    memset(&rvi, 0, sizeof rvi);

    memset(&gc, 0, sizeof gc);
    gc.flags = VGCF_in_kernel | VGCF_online;
    gc.ctrlreg[3] = rdcr3();
    asm(" mov %%ds,%0" : "=r"(gc.user_regs.ds));
    asm(" mov %%es,%0" : "=r"(gc.user_regs.es));
    asm(" mov %%fs,%0" : "=r"(gc.user_regs.fs));
    asm(" mov %%gs,%0" : "=r"(gc.user_regs.gs));
    asm(" mov %%ss,%0" : "=r"(gc.user_regs.ss));
    asm(" mov %%cs,%0" : "=r"(gc.user_regs.cs));
    gc.user_regs.rip = (uint64_t)run_vcpu;
    gc.kernel_ss = gc.user_regs.ss;

    for (id = 1; id < MAXVCPU; id++) {
	if (id >= XEN_LEGACY_MAX_VCPUS) {
	    if (vicnt <= 0) {
		fvi = malloc(PAGESIZE);
		rvi.mfn = vaddr2mfn(fvi);
		vicnt = PAGESIZE / sizeof *fvi;
	    }
	    vi = fvi + --vicnt;
	    rvi.offset = (uint64_t)vi & (PAGESIZE - 1);
	    if (hypercall_vcpu_op(VCPUOP_register_vcpu_info, id, &rvi) != 0) {
		if (vicnt == PAGESIZE / sizeof *fvi - 1)
		    free(fvi);
		break;
	    }
	} else
	    vi = shared_info->vcpu_info + id;
	/* No interrupts ever! */
	vi->evtchn_upcall_mask = 1;

	/*
	 * For now we allocate only one page for the stack.
	 */
	gc.user_regs.rsp = (uint64_t)malloc(PAGESIZE) + PAGESIZE - 8;
	gc.kernel_sp = gc.user_regs.rsp;

	/*
	 * Why do we need to down the vcpu before upping it???
	 */
	if (hypercall_vcpu_op(VCPUOP_initialise, id, &gc) != 0
	    || hypercall_vcpu_op(VCPUOP_down, id, 0) != 0
	    || hypercall_vcpu_op(VCPUOP_up, id, 0) != 0) {
	    /*
	     * additional error handling???
	     * e.g. are we allowed to free the vcpu_info here???
	     */
	    break;
	}
    }
    printf("Using %d VCPUs\n", id);
}
