/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <MiXen.h>
#include <atomic.h>
#include <memory.h>
#include <string.h>
#include <systm.h>

#include <event_channel.h>
#include <io/console.h>

static volatile struct xencons_interface *xencons;

#define	XCO_IDX(idx)	MASK_XENCONS_IDX((idx), xencons->out)

/*
 * Flush the console ring buffer.
 * For now, we do it synchronously, i.e. this routine returns only
 * after the backend has consumed all the data.
 */
static void
console_flush(void)
{
    evtchn_send_t send;

    send.port = start_info->console.domU.evtchn;
    hypercall_event_channel_op(EVTCHNOP_send, &send);
    while (XCO_IDX(xencons->out_prod - xencons->out_cons) > 0)
	pollevents(start_info->console.domU.evtchn);
}

/*
 * Write the len bytes from buf to the console.
 * For now this blocks until all bytes are delivered.
 */
static void
console_out(char *buf, size_t len)
{
    unsigned long l, o;
    static uint32_t console_lock;

    if (!xencons)
	return;

    while (atomic_cas_32(&console_lock, 0, 1) != 0)
	yield();

    while (len > 0) {
	o = XCO_IDX(xencons->out_prod);
	l = XCO_IDX(xencons->out_cons - o);
	if (l == 0) {
	    /* prod == cons means the ring is empty, not full */
	    l = sizeof xencons->out;
	}
	if (l > len)
	    l = len;
	if (l > sizeof xencons->out - o)
	    l = sizeof xencons->out - o;
	if (l < 1) {
	    /* no room in buffer, tell the other side */
	    console_flush();
	    continue;
	}
	memcpy((char *)xencons->out + o, buf, l);
	buf += l;
	len -= l;
	xencons->out_prod += l;
    }
    if (XCO_IDX(xencons->out_prod - xencons->out_cons))
	console_flush();

    console_lock = 0;
}

/*
 * To cite a famous comment:
 * This is not for chit-chat!
 */
void
panic(const char *msg)
{
    printf("%s\n", msg);
    for (;;)
	halt();
}

void
failure(const char *func, const char *msg)
{
    printf("%s: ", func);
    panic(msg);
}

void
console_write(void *arg0, const char *s, size_t len)
{
    char buf[256], *dp;

    for (dp = buf; len > 0;) {
	switch (*s) {
	case '\n':
	    if (dp + 2 >= buf + sizeof buf) {
		console_out(buf, dp - buf);
		dp = buf;
	    }
	    *dp++ = '\r';
	    /* FALLTHRU */
	default:
	    *dp++ = *s++;
	    len--;
	    break;
	}
	if (dp >= buf + sizeof buf) {
	    console_out(buf, dp - buf);
	    dp = buf;
	}
    }
    if (dp > buf)
	console_out(buf, dp - buf);
}

#include <vcpu.h>
static void
console_intr(void *arg)
{
    /*
     * For now, we just consume any input delivered to us.
     */
    xencons->in_cons = xencons->in_prod;
}

void
init_console(void)
{
    /*
     * The initial mapping has 1 page for xenstore and 1 page
     * for the console immediatly behind the start_info.
     */
    xencons = (void *)((char *)start_info + 2 * PAGESIZE);

    install_handler(start_info->console.domU.evtchn, &console_intr, 0);
}
