/*
 * Copyright (c) 2011 Wolfgang Solfrank.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the  
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Wolfgang Solfrank.
 * 4. The name of TooLs GmbH may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY TOOLS GMBH ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL TOOLS GMBH BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * xenstore message handling.
 */
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <systm.h>

#include <MiXen.h>
#include <memory.h>

#include <event_channel.h>
#include <io/protocols.h>
#include <io/xenbus.h>
#include <io/xs_wire.h>

static struct xenstore_domain_interface *xsdi;

#ifdef	DEBUG
char *msgtypes[] = {
    "XS_DEBUG",
    "XS_DIRECTORY",
    "XS_READ",
    "XS_GET_PERMS",
    "XS_WATCH",
    "XS_UNWATCH",
    "XS_TRANSACTION_START",
    "XS_TRANSACTION_END",
    "XS_INTRODUCE",
    "XS_RELEASE",
    "XS_GET_DOMAIN_PATH",
    "XS_WRITE",
    "XS_MKDIR",
    "XS_RM",
    "XS_SET_PERMS",
    "XS_WATCH_EVENT",
    "XS_ERROR",
    "XS_IS_DOMAIN_INTRODUCED",
    "XS_RESUME",
    "XS_SET_TARGET",
    "XS_RESTRICT"
};
#endif

#define	NOUTMSG	(NGUESTS * 8)
#define	NINMSG	(NGUESTS * 2)

static struct xsd_sockmsg *outmsgs[NOUTMSG];
static size_t outprod, outcons;
#define	OUTIDX(idx)	((idx) % NOUTMSG)

static struct {
    uint32_t rid;
    uint32_t tid;
    void (*done)(struct xsd_sockmsg *, void *);
    void *arg;
} indone[NINMSG];

static struct backend {
    uint32_t flags;
#define	BK_DYING		0x01	/* connection is dying */
#define	BK_ERROR		0x02	/* set hotplug-status to error */
#define	BK_BACK_TX		0x04	/* backend transaction in progress */
#define	BK_BACK_TX_AGAIN	0x08	/* repeat b_tx after completion */
#define	BK_PROTO_32		0x10	/* frontend uses x86_32-abi */
    uint32_t back_tx;
    uint32_t front_tx;
    enum xenbus_state state;
    size_t blen;
    char backend[XENSTORE_REL_PATH_MAX];
    char frontend[XENSTORE_ABS_PATH_MAX];
    void *arg;
    domid_t dom;
    grant_ref_t ring;
    evtchn_port_t event;
} backends[NGUESTS];

#define	errcmp(s, err)	(memcmp(((struct xsd_sockmsg *)(s)) + 1, \
			 (err), sizeof(err) - 1))

static struct xsd_sockmsg *
msgalloc(void)
{
    return malloc(PAGESIZE);
}

static void
msgfree(struct xsd_sockmsg *msg)
{
    free(msg);
}

/*
 * Signal the other side that there is work to do or at least some
 * more space available in the ringbuffer.
 */
static void
xs_signal(void)
{
    evtchn_send_t send;

    xen_wmb();
    send.port = start_info->store_evtchn;
    ASSERT(hypercall_event_channel_op(EVTCHNOP_send, &send) == 0);
}

#ifdef	DEBUG
static void
dumpmsg(struct xsd_sockmsg *msg, char *dir)
{
    char *s, *e;

    DPRINTF(DBG_XS, "xs %s %s(%u, %u): ",
	    dir,
	    msg->type < __arraycount(msgtypes)
		? msgtypes[msg->type]
		: "XS_???",
	    msg->req_id,
	    msg->tx_id);
    for (s = (char *)(msg + 1), e = s + msg->len; s < e; s++)
	if (*s)
	    DPRINTF(DBG_XS, "%c", *s);
	else
	    DPRINTF(DBG_XS, " ");
    DPRINTF(DBG_XS, "\n");
}
#endif

/*
 * Stuff data into output ringbuffer.
 */
static void
outchunk(void)
{
    static struct xsd_sockmsg *msg;
    static char *output;
    static size_t outlen;
    size_t cnt;
    int done = 0;

    while (1) {
	/* Ensure that output and outlen are pointing to some data. */
	if (outlen == 0) {
	    output = 0;
	    if (msg) {
		/* Free output buffer which was already sent. */
		msgfree(msg);
		msg = 0;
	    }

	    /* Try to get another output message. */
	    if (OUTIDX(outprod - outcons) > 0) {
		msg = outmsgs[OUTIDX(outcons++)];
#ifdef	DEBUG
		dumpmsg(msg, "out");
#endif
		output = (char *)msg;
		outlen = sizeof *msg + msg->len;
		ASSERT(outlen > 0);
	    }

	    if (outlen == 0) {
		/*
		 * No more data available.
		 * If we actually did output something, signal the other side.
		 */
		if (done)
		    xs_signal();
		return;
	    }
	}

	/* Determine free area in ringbuffer. */
	switch (cnt = MASK_XENSTORE_IDX(xsdi->req_prod - xsdi->req_cons)) {
	case 0:			/* Means empty, not full */
	    cnt = sizeof xsdi->req - 1;
	    break;
	case 1:			/* So this has to mean full */
	    xs_signal();
	    return;
	}

	/* Ok, let's output some chunk of data. */
	done = 1;
	if (cnt > outlen)
	    cnt = outlen;
	if (cnt > sizeof xsdi->req - MASK_XENSTORE_IDX(xsdi->req_prod))
	    cnt = sizeof xsdi->req - MASK_XENSTORE_IDX(xsdi->req_prod);
	memcpy(xsdi->req + MASK_XENSTORE_IDX(xsdi->req_prod), output, cnt);
	xsdi->req_prod += cnt;
	output += cnt;
	outlen -= cnt;
    }
}

/*
 * Put some message into the output queue.
 */
static void
xs_send(struct xsd_sockmsg *msg)
{
    outmsgs[OUTIDX(outprod++)] = msg;
    ASSERT(OUTIDX(outprod - outcons) != 0);
    outchunk();
}

static void watch_event(struct xsd_sockmsg *);

/*
 * This gets called when a complete message was received from the ringbuffer.
 */
static void
input_done(struct xsd_sockmsg *msg)
{
    size_t i;
    void (*done)(struct xsd_sockmsg *, void *);
    void *arg;

#ifdef	DEBUG
    dumpmsg(msg, "in");
#endif

    /*
     * For now, we throw away any unexpected input.
     */
    if (msg->type == XS_WATCH_EVENT
	&& msg->req_id == 0 && msg->tx_id == 0)
	watch_event(msg);
    else {
	/*
	 * Look whether we were waiting for this response.
	 */
	for (i = 0; i < __arraycount(indone); i++)
	    if (indone[i].rid == msg->req_id
		&& indone[i].tid == msg->tx_id) {
		done = indone[i].done;
		arg = indone[i].arg;
		indone[i].rid = 0;
		done(msg, arg);
		break;
	    }
    }
}

/*
 * Indicate that we are waiting for some response.
 */
static void
handle_input(struct xsd_sockmsg *msg,
	void (*done)(struct xsd_sockmsg *, void *), void *arg)
{
    size_t i;

    for (i = 0; i < __arraycount(indone); i++)
	if (indone[i].rid == 0) {
	    indone[i].rid = msg->req_id;
	    indone[i].tid = msg->tx_id;
	    indone[i].done = done;
	    indone[i].arg = arg;
	    return;
	}
    panic("no available slot for indone");
}

/*
 * "Interrupt" handler for xenstore.
 */
static void
xenstore_intr(void *arg)
{
    static union {
	struct xsd_sockmsg msg;
	char buf[PAGESIZE];
    } mbuf;
    static char *input = mbuf.buf;
    size_t cnt, need;
    int done = 0;

    /* First, try to output data the was delayed due to ringbuffer full. */
    outchunk();

    /* Now handle any available input. */
    while ((cnt = MASK_XENSTORE_IDX(xsdi->rsp_prod - xsdi->rsp_cons)) > 0) {
	done = 1;

	need = mbuf.buf + sizeof mbuf.msg + mbuf.msg.len - input;
	if (cnt > need)
	    cnt = need;
	if (cnt > sizeof xsdi->rsp - MASK_XENSTORE_IDX(xsdi->rsp_cons))
	    cnt = sizeof xsdi->rsp - MASK_XENSTORE_IDX(xsdi->rsp_cons);
	memcpy(input, xsdi->rsp + MASK_XENSTORE_IDX(xsdi->rsp_cons), cnt);
	xsdi->rsp_cons += cnt;
	input += cnt;
	ASSERT(sizeof mbuf.msg + mbuf.msg.len < PAGESIZE);
	if (input - mbuf.buf >= sizeof mbuf.msg + mbuf.msg.len) {
	    input_done((void *)mbuf.buf);
	    mbuf.msg.len = 0;
	    input = mbuf.buf;
	}
    }
    if (done)
	xs_signal();
}

/*
 * Return some id for a request that's not in use.
 * We rely on the number space to be large enough to not wrap while some
 * id is in use.
 */
static uint32_t
nextid(void)
{
    static uint32_t id;

    if (++id == 0)
	id++;
    return id;
}

/*
 * Construct a message and output it to xenstore.
 * Possible fmt characters:
 * T:	The next arg is an uint32_t with the transaction id.
 * u:	The next arg is an uint32_t that gets sprinted to the buffer.
 * U:	The next arg is an uint64_t that gets sprinted to the buffer.
 * s:	The next arg is a '\0'-terminated string.
 * S:	The next arg is a string, but the terminating '\0' is not copied.
 * F:	The next 2 args are the continuation function and arg.
 *	If no 'F' is given, any response is ignored.
 * A:	Must be last in list. Signifies a recursion in argument handling,
 *	meaning that we fetch a new format and va_list from the arguments.
 */
void
xs_msg(uint32_t type, const char *fmt, ...)
{
    struct xsd_sockmsg *msg;
    char *cp, *end, *s;
    size_t l;
    uint64_t u;
#ifdef	STDARG_FIXED
    va_list ap, ap1;
#else
    va_list ap;
#endif
    void (*done)(struct xsd_sockmsg *, void *) = 0;
    void *arg = 0;

    msg = msgalloc();
    msg->type = type;
    msg->req_id = nextid();
    msg->tx_id = 0;
    cp = (char *)(msg + 1);
    end = (char *)msg + PAGESIZE;
    va_start(ap, fmt);
    while (*fmt)
	switch (*fmt++) {
	case 'T':
	    msg->tx_id = va_arg(ap, uint32_t);
	    break;
	case 'u':
	    u = va_arg(ap, uint32_t);
	    l = snprintf(cp, (char *)msg + PAGESIZE - cp, "%u", u);
	    ASSERT(cp + l <= end);
	    cp += l;
	    break;
	case 'U':
	    u = va_arg(ap, uint64_t);
	    l = snprintf(cp, (char *)msg + PAGESIZE - cp, "%lu", u);
	    ASSERT(cp + l <= end);
	    cp += l;
	    break;
	case 's':
	    s = va_arg(ap, char *);
	    l = strlen(s) + 1;
	    ASSERT(cp + l <= end);
	    memcpy(cp, s, l);
	    cp += l;
	    break;
	case 'S':
	    s = va_arg(ap, char *);
	    l = strlen(s);
	    ASSERT(cp + l < end);
	    memcpy(cp, s, l);
	    cp += l;
	    break;
	case 'F':
	    done = va_arg(ap, void (*)(struct xsd_sockmsg *, void *));
	    arg = va_arg(ap, void *);
	    break;
#ifdef	STDARG_FIXED
	case 'A':
	    if (!*fmt) {
		fmt = va_arg(ap, char *);
		va_copy(ap1, va_arg(ap, va_list));
		va_end(ap);
		va_copy(ap, ap1);
		va_end(ap1);
		break;
	    }
#endif
	default:
	    panic("invalid xs_msg parameter");
	}
    va_end(ap);
    msg->len = cp - (char *)(msg + 1);
    if (done)
	handle_input(msg, done, arg);
    xs_send(msg);
}

#ifdef	STDARG_FIXED
/*
 * Used by external routines to write values into nodes.
 * Currently used only by backend_setparams.
 *	arg is an opaque pointer given as argument to backend_setparams.
 *	node is the name of the node in xenstore (without the leading backend
 *		path) into which to write the value.
 *	fmt and args give the values to be written into the node.
 *		Note: not really checked.  Supported characters are "uUsS"
 *		handled as with xs_msg above.
 */
void
xs_write(void *arg, const char *node, const char *fmt, ...)
{
    struct backend *bk = arg;
    va_list ap;

    va_start(ap, fmt);
    xs_msg(XS_WRITE, "TSsA", bk->back_tx, bk->backend, node, fmt, ap);
    va_end(ap);
}
#else
/*
 * The above routine is currently only used to write some unsigned integer
 * value, sometimes 32 bit, sometimes 64 bit into some node.
 * So for now, we get away with always writing a uint64_t into the node.
 */
void
xs_uint64(void *arg, const char *node, uint64_t val)
{
    struct backend *bk = arg;

    xs_msg(XS_WRITE, "TSsU", bk->back_tx, bk->backend, node, val);
}
#endif

static int bk_dying(struct backend *);
static void bk_unwatch(struct xsd_sockmsg *, void *);

/*
 * Check whether a backend is dying.
 * If so, free the backend iff there is no I/O going on for it.
 */
static int
bk_dying(struct backend *bk)
{
    int i;

    if (!(bk->flags & BK_DYING))
	return 0;

    /*
     * Remove the frontend watch if it is still in use.
     */
    if (bk->frontend[0]) {
	xs_msg(XS_UNWATCH, "sUsF", bk->frontend, bk, "", bk_unwatch, bk);
	bk->frontend[0] = 0;
	return 1;
    }

    /*
     * We check that there is no onging transaction in indone
     * referring to this bk.  If there is one, we get called
     * again for this backend.
     */
    for (i = 0; i < __arraycount(indone); i++)
	if (indone[i].rid && indone[i].arg == bk)
	    return 1;

    if (bk->backend[0] && (bk->flags & BK_ERROR))
	xs_msg(XS_WRITE, "SsS", bk->backend, "/hotplug-status", "error");

    backend_destroy(bk->arg);
    memset(bk, 0, sizeof *bk);
    bk->ring = (grant_ref_t)-1;
    bk->dom = (domid_t)-1;

    return 1;
}

/*
 * Calback here when xenstore acks the UNWATCH request for the frontend.
 */
static void
bk_unwatch(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_UNWATCH)
	panic("bk_unwatch");

    bk_dying(bk);
}

static void bk_getstate(struct xsd_sockmsg *, void *);
static void bk_back_state(struct xsd_sockmsg *, void *);
static void bk_back_end(struct xsd_sockmsg *, void *);
static void bk_back_txok(struct xsd_sockmsg *, void *);

/*
 * Read the backend state in a transaction.
 */
static void
bk_getstate(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_TRANSACTION_START)
	panic("bk_getstate");

    ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
    ((char *)(msg + 1))[msg->len] = 0;
    bk->back_tx = atoi((char *)(msg + 1));

    xs_msg(XS_READ, "TSsF",
	bk->back_tx, bk->backend, "/state", bk_back_state, arg);
}

/*
 * Handle the current backend state.
 */
static void
bk_back_state(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;
    enum xenbus_state state;

    switch (msg->type) {
    case XS_ERROR:
	if (errcmp(msg, "ENOENT"))
	    panic("bk_back_state error");
	bk->flags |= BK_DYING;
	bk_dying(bk);
	break;
    case XS_READ:
	ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
	((char *)(msg + 1))[msg->len] = 0;
	bk->state = atoi((char *)(msg + 1));

	switch (bk->state) {
	case XenbusStateInitialising:
	    /*
	     * Set parameters for the frontend.
	     */
	    if (!(bk->arg = backend_setparams(bk))) {
		state = XenbusStateClosed;
		xs_msg(XS_WRITE, "TSsS",
		    bk->back_tx, bk->backend, "hotplug-status", "error");
	    } else
		/*
		 * If we didn't get the frontend data yet,
		 * we have to wait for it.
		 */
		if (bk->event == 0)
		    state = XenbusStateInitWait;
		else
		    state = XenbusStateInitialised;
	    break;
	case XenbusStateInitialised:
	    if (bk->event > 0) {
		/*
		 * Got all neccessary data, so launch the service.
		 */
		state = XenbusStateConnected;
		backend_setup(bk->arg, bk->dom,
		    bk->ring, bk->event, bk->flags & BK_PROTO_32);
		break;
	    }
	    goto leave_alone;
	case XenbusStateInitWait:
	    /*
	     * If in the meantime, we got the backend data, go ahead.
	     */
	    if (bk->event > 0) {
		state = XenbusStateInitialised;
		break;
	    }
	case XenbusStateConnected:
	case XenbusStateClosed:
 leave_alone:
	    /*
	     * Nothing to do here, just stop the transaction.
	     */
	    xs_msg(XS_TRANSACTION_END, "Ts", bk->back_tx, "F");
	    bk->flags &= ~BK_BACK_TX;
	    if (!bk_dying(bk) && (bk->flags & BK_BACK_TX_AGAIN)) {
		bk->flags &= ~BK_BACK_TX_AGAIN;
		bk->flags |= BK_BACK_TX;
		xs_msg(XS_TRANSACTION_START, "sF", "", bk_getstate, arg);
	    }
	    return;
	default:
	    state = XenbusStateClosed;
	    break;
	}

	/*
	 * Write out new backend state.
	 */
	xs_msg(XS_WRITE, "TSsuF",
	    bk->back_tx, bk->backend, "/state", state, bk_back_end, arg);
	break;
    default:
	panic("bk_back_state type");
    }
}

/*
 * After writing the new backend state, finish the transaction.
 */
static void
bk_back_end(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_WRITE)
	panic("bk_back_end");

    xs_msg(XS_TRANSACTION_END, "TsF", bk->back_tx, "T", bk_back_txok, arg);
}

/*
 * Check whether we have to restart the transaction.
 */
static void
bk_back_txok(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    switch (msg->type) {
    case XS_ERROR:
	if (errcmp(msg, "EAGAIN"))
	    panic("bk_back_txok error");
	if (!bk_dying(bk)) {
	    /*
	     * Let's try again!
	     */
	    bk->flags &= ~BK_BACK_TX_AGAIN;
	    xs_msg(XS_TRANSACTION_START, "sF", "", bk_getstate, arg);
	}
	break;
    case XS_TRANSACTION_END:
	bk->flags &= ~BK_BACK_TX;
	if (!bk_dying(bk) && (bk->flags & BK_BACK_TX_AGAIN)) {
	    bk->flags &= ~BK_BACK_TX_AGAIN;
	    bk->flags |= BK_BACK_TX;
	    xs_msg(XS_TRANSACTION_START, "sF", "", bk_getstate, arg);
	}
	break;
    default:
	panic("bk_back_txok type");
	break;
    }
}

/*
 * Got the frontend path from xenstore.
 */
static void
bk_read_front(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    switch (msg->type) {
    case XS_READ:
	ASSERT(msg->len < XENSTORE_ABS_PATH_MAX);
	((char *)(msg + 1))[msg->len] = 0;
	strcpy(bk->frontend, (char *)(msg + 1));

	/* Here, the token requires a trailing '\0'! */
	xs_msg(XS_WATCH, "sUs", bk->frontend, arg, "");
	break;

    case XS_ERROR:
	if (!errcmp(msg, "ENOENT")) {
	    /*
	     * Looks like we were too slow in reading this entry.
	     */
	    bk->flags |= BK_DYING | BK_ERROR;
	    bk_dying(bk);
	    break;
	}
    default:
	panic("bk_read_front");
    }
}

static void bk_front_state(struct xsd_sockmsg *, void *);
static void bk_ring(struct xsd_sockmsg *, void *);
static void bk_event(struct xsd_sockmsg *, void *);
static void bk_proto(struct xsd_sockmsg *, void *);

/*
 * Check the current frontend state.
 */
static void
bk_front_state(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;
    enum xenbus_state state;

    switch (msg->type) {
    case XS_READ:
	ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
	((char *)(msg + 1))[msg->len] = 0;
	state = atoi((char *)(msg + 1));

	if (state == XenbusStateInitialised
	    && bk->ring == (grant_ref_t)-1)
	    xs_msg(XS_READ, "TSsF",
		bk->front_tx, bk->frontend, "/ring-ref", bk_ring, arg);
	break;
    case XS_ERROR:
	if (!errcmp(msg, "ENOENT") || !errcmp(msg, "EACCES")) {
	    /*
	     * This means that the frontend gets destroyed!
	     */
	    bk->flags |= BK_DYING;
	    bk_dying(bk);
	    break;
	}
    default:
	panic("bk_front_state");
    }
}

/*
 * Got the ringbuffer page from xenstore for front-/backend communication.
 */
static void
bk_ring(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_READ)
	panic("bk_ring");

    ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
    ((char *)(msg + 1))[msg->len] = 0;
    bk->ring = atoi((char *)(msg + 1));
    ASSERT(bk->ring != (grant_ref_t)-1);

    xs_msg(XS_READ, "TSsF",
	bk->front_tx, bk->frontend, "/event-channel", bk_event, arg);
}

/*
 * Got the event channel for front-/backend communication.
 */
static void
bk_event(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_READ)
	panic("bk_event");
    ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
    ((char *)(msg + 1))[msg->len] = 0;
    bk->event = atoi((char *)(msg + 1));
    ASSERT(bk->event > 0);

    xs_msg(XS_READ, "TSsF",
	bk->front_tx, bk->frontend, "/protocol", bk_proto, arg);
}

/*
 * Got the protocol type which the frontend uses.
 */
static void
bk_proto(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_READ)
	panic("bk_proto");
    ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
    ((char *)(msg + 1))[msg->len] = 0;
    if (!strcmp((char *)(msg + 1), XEN_IO_PROTO_ABI_X86_32))
	bk->flags |= BK_PROTO_32;
    else if (strcmp((char *)(msg + 1), XEN_IO_PROTO_ABI_X86_64))
	panic((char *)(msg + 1));

    /*
     * Since this may be the last data needed to go,
     * trigger a read of the backend state.
     */
    if (bk->flags & BK_BACK_TX)
	bk->flags |= BK_BACK_TX_AGAIN;
    else {
	bk->flags |= BK_BACK_TX;
	xs_msg(XS_TRANSACTION_START, "sF", "", bk_getstate, bk);
    }
}

/*
 * This probably isn't neccessary!
 * We got a watch event for some backend we already knew about,
 * we expect that since this is going to be destroyed, the read of
 * the backend state fails.
 * We could also set BK_DYING right where we initiate this read below.
 */
static void
bk_remove(struct xsd_sockmsg *msg, void *arg)
{
    struct backend *bk = arg;

    if (msg->type != XS_ERROR || errcmp(msg, "ENOENT"))
	panic("bk_remove");
    bk->flags |= BK_DYING;
    bk_dying(bk);
}

/*
 * Some frontend WATCH event arrived.
 */
static void
watch_frontend(struct xsd_sockmsg *msg)
{
    char *path = (char *)(msg + 1);
    char *cp;
    struct backend *bk;

    ASSERT(msg->len < PAGESIZE - sizeof *msg - 1);
    path[msg->len] = 0;
    cp = path + strlen(path) + 1;
    ASSERT(cp < (char *)msg + PAGESIZE);
    bk = (struct backend *)atol(cp);

    for (cp -= 1; cp >= path;)
	if (*--cp == '/')
	    break;
    /*
     * Since we only come here if path starts with a '/',
     * no need here for an ASSERT.
     */
    /*
     * If we are signalled a change in the state,
     * try to read the state, if no other transaction is going on.
     * Since we may have been to slow setting up the watch, we also
     * have to do this, if we are getting just the initial watch event.
     */
    if (!strcmp(cp, "/state") || !strcmp(path, bk->frontend))
	if (!bk_dying(bk))
	    xs_msg(XS_READ, "SsF", bk->frontend, "/state", bk_front_state, bk);
}

/*
 * Some watch event arrived.
 */
static void
watch_event(struct xsd_sockmsg *msg)
{
    struct backend *bk = 0, *fb = 0;
    size_t slashes, i;
    char *path = (char *)(msg + 1);
    char *cp;

    /*
     * Since all backend paths are relative,
     * an absolute path indicates a frontend event.
     */
    if (path[0] == '/') {
	watch_frontend(msg);
	return;
    }

    ASSERT(!memcmp(path, BACKEND, sizeof BACKEND - 1));

    /*
     * Find either the already existing backend path
     * or an empty slot in the backend table.
     */
    for (i = 0; i < __arraycount(backends); i++) {
	if (backends[i].backend[0] == 0)
	    fb = &backends[i];
	else if (!memcmp(backends[i].backend, path, backends[i].blen)
	    && (path[backends[i].blen] == 0 || path[backends[i].blen] == '/'))
	    bk = &backends[i];

	if (fb && bk)
	    break;
    }

    /*
     * Determine whether there is something after
     * <backend>/<domid>/<devicenode>
     */
    slashes = 0;
    for (cp = path + sizeof BACKEND - 1; *cp;)
	if (*cp++ == '/')
	    if (++slashes > 2)
		break;

    if (slashes == 2) {
	/*
	 * No, this is just the plain path,
	 * so xenstore is either telling us the creation or the
	 * destruction of some object.
	 */

	if (bk) {
	    /*
	     * This backend path exists already.
	     * This means, the backend is probably going to be destroyed.
	     * We try to read the state and free the entry on error.
	     */
	    xs_msg(XS_READ, "SsF", bk->backend, "/state", bk_remove, bk);
	    return;
	}

	if (fb == 0) {
	    /*
	     * No free slot, deny the request.
	     */

	    /*
	     * It looks like the write to the hotplug-status shouldn't include
	     * the trailing '\0'.
	     */
	    xs_msg(XS_WRITE, "SsS", path, "/hotplug-status", "busy");
	    return;
	}

	strcpy(fb->backend, path);
	fb->blen = strlen(path);
	fb->dom = atoi(path + sizeof BACKEND);

	/*
	 * See above about trailing '\0'.
	 */
	xs_msg(XS_WRITE, "SsS", path, "/hotplug-status", "connected");
    } else {
	/*
	 * If we didn't find the backend path above, just ignore the request.
	 */
	if (!bk)
	    return;

	/*
	 * We are interested in any change in state and frontend path.
	 */
	if (!strcmp(cp, "state"))
	    if (bk->flags & BK_BACK_TX)
		bk->flags |= BK_BACK_TX_AGAIN;
	    else {
		bk->flags |= BK_BACK_TX;
		xs_msg(XS_TRANSACTION_START, "sF", "", bk_getstate, bk);
	    }

	else if (!strcmp(cp, "frontend"))
	    xs_msg(XS_READ, "sF", path, bk_read_front, bk);
    }
}

/*
 * Initialize xenstore handling.
 */
void
init_xenstore(void)
{
    int idx;

    /*
     * The initial mapping has 1 page for xenstore and 1 page
     * for the console immediatly behind the start_info.
     */
    xsdi = (void *)((char *)start_info + PAGESIZE);

    for (idx = 0; idx < NGUESTS; idx++) {
	backends[idx].dom = (domid_t)-1;
	backends[idx].ring = (grant_ref_t)-1;
    }

    install_handler(start_info->store_evtchn, xenstore_intr, 0);

    /*
     * According to xenstore.txt, there should be a trailing reserved
     * value in this message, however, it doesn't look that's really
     * neccessary.
     * For now, we are not interested in the response to this.
     */
    xs_msg(XS_WATCH, "ss", BACKEND, "MiXen");
}
